<?php
include_once("utils/bdUtils.php");

class oauthRS {

    protected $resource;
    protected $keys;
    protected $scope;
    protected $error;
    protected $bearertoken_request;
    protected $accesstoken_request;
    protected $token;
    protected $extra;
    protected $isExpired;
    protected $activeDebug;

    public function __construct() {
        $this->resource=null;
        $this->scope = null;
        $this->error=null;
        $this->resource = null;
        $this->accesstoken_request = null;
        $this->bearertoken_request=null;
        $this->extra=array();
        $this->isExpired=false;
        $this->getKeys();
        $this->deactiveDebug();
    }

    /**
     * Function that manage the request of the client and give a response with the request in a valid case and
     * in the case of an error, the response will be an HTTP 400 or 401 (depending on the error)
     */
    public function manageRequest() {
        if($this->activeDebug) error_log("oauthRS.manageRequest");
        $dev= false;
        $host = (isset($_SERVER['HTTP_X_FORWARDED_FOR'])) ? $_SERVER['HTTP_X_FORWARDED_FOR'] : $_SERVER['REMOTE_ADDR'];
        if (array_key_exists("CONTENT_TYPE", $_SERVER)
                && (strcmp($_SERVER['CONTENT_TYPE'],"application/x-www-form-urlencoded")==0)
                && isset($_POST)
                && sizeof($_POST)>0
        ) {
            $dev = $this->manageResourceRequest("form",$_POST,$_SERVER['SERVER_NAME'].$_SERVER['REQUEST_URI'],
                    $_SERVER['SERVER_PORT'], $host, $_SERVER['HTTPS']);

        }else if(isset($_GET)&& (sizeof($_GET)>0)) {
            $dev = $this->manageResourceRequest("get", $_GET, $_SERVER['SERVER_NAME'].$_SERVER['REQUEST_URI'],
                    $_SERVER['SERVER_PORT'], $host,$_SERVER['HTTPS']);

        }else {
            $dev = $this->manageResourceRequest("header", apache_request_headers(),$_SERVER['SERVER_NAME'].$_SERVER['REQUEST_URI'],
                    $_SERVER['SERVER_PORT'], $host,$_SERVER['HTTPS']);
        }
        if($dev) {
            echo $this->getResource();
        }else {
            $this->getMessageError();
        }
    }

    /**
     * Managin the request, checking if it's a valid one,
     * @param <String> $type of the request (x-www-form-urlencoded, an http auth header, or a get request)
     * @param <Array>|<String> $data Parameters of the request
     * @param <String> $servername name of the server
     * @param <int> $serverport port of the server
     * @param <String> $hostname IP of the client
     * @param <String> $httpsOn "on" if the connection is https
     * @return <bool>  True if the request is a valid one
     */
    private function manageResourceRequest($type,$data, $servername, $serverport, $hostname, $httpsOn) {
        if($this->activeDebug) error_log("oauthRS.manageResourceRequest.");
        $dev = false;
        $continue = false;
        if(strcmp($type, "header")==0) {
            $continue = $this->getAuthorizationHeader($data);
        }else if(strcmp($type, "form")==0) {
            $continue = $this->getPostData($data);
        }else if(strcmp($type,"get")==0) {
            $continue = $this->getGetData($data);
        }
        if($continue) {
            if((null !=$this->bearertoken_request) && $this->isValidBearerToken($hostname, $httpsOn)) {
                $dev =true;
            }else if((null !=$this->accesstoken_request) && $this->isValidAccessToken($servername, $serverport, $hostname, $type, $httpsOn)) {
                $dev =true;
            }else {
                $this->error .= "Invalid Token.";
            }
        }else {
            $this->error .= "Invalid Request.";
        }
        return $dev;
    }

    /**
     *Function that checks if the token given in the request is a valid one.
     * @param <type> $token Access or bearer token
     * @param <type> $hostname IP of the client
     * @param <type> $httpsOn "on" if the request was in https
     * @return <bool>  true if the token is a valid one
     */
    private function isValidToken($token, $hostname, $httpsOn) {
        if($this->activeDebug)error_log("oauthRS.isValidToken" );
        $dev =false;
        if(!array_key_exists($hostname, $this->keys)) {
            $this->error = "invalid_client_credentials";
        }else {
            $key = $this->keys[$hostname];
            //Les llega un X:Y, donde
            // X = hmac(Y, hostname)
            // Y = hmac(M,key):M
            // M = b64e(spuc):b64e(scope):b64e(lifetime)
            $pos1 = strpos($token, ":");
            $hashIp = substr($token, 0, $pos1);
            $menIp = substr($token, $pos1+1);
            if(0==strcmp($hashIp, hash_hmac("sha256",$menIp,$hostname))) {
                $array = explode(":", $menIp);
                $m = $array[0];
                $scope = base64_decode($array[2]);
                $this->scope = $scope;
                $time = base64_decode($array[3]);
                if($time>microtime(true)) {
                    $pos = strpos($menIp,":");
                    $message =  substr($menIp, $pos+1);
                    if(0==strcmp($m, hash_hmac("sha256",$message,$key))) {
                        $dev = true;
                    }
                }else {
                    $this->error.="authorization_expired";
                    $this->isExpired = true;
                }
            }
        }
        return $dev;
    }

    /**
     *Function that checks if the access token given in the request is a valid one.
     * @param <String> $servername name of the server
     * @param <int> $serverport port of the server
     * @param <String> $hostname IP of the client
     * @param <String> $method Signature method. Supported values: "hmac-sha256".
     * @param <String> $httpsOn "on" if the connection is https
     * @return <bool>  true if the token is a valid one
     */
    private function isValidAccessToken($servername, $serverport, $hostname, $method, $httpsOn) {
        if($this->activeDebug) error_log("oauthRS.isValidAccessToken");
        $dev = false;
        $token = $this->accesstoken_request['token'];
        $signature =  $this->accesstoken_request['signature'];
        $algorithm =  $this->accesstoken_request['algorithm'];
        $nonce =  $this->accesstoken_request['nonce'];
        $timestamp =  $this->accesstoken_request['timestamp'];

        //comprobar que el algoritmo es sha256
        if(0!=strcmp("hmac-sha256", $algorithm)) {
            $this->error .= "Invalid cipher algorithm. ";
        }else {
            //validar token
            if(!$this->isValidToken($token, $hostname, $httpsOn)) {
                $this->error .= "Invalid Token. ";
            }else {
                //comprobar que la firma es la adecuada
                if(strcmp($httpsOn,"on")==0) {
                    $servername="https://".$servername;
                }else {
                    $servername="http://".$servername;
                }
                $message = $timestamp.",".$nonce.",".$algorithm.",".$method.",".$servername.":".$serverport.",".$this->scope;
                if(0!=strcmp($signature, hash_hmac("sha256", $message, $this->key))) {
                    $this->error .= "Invalid request signature. ";
                }else {
                    //comprobar que el nonce no se repite para ese cliente
                    if($this->existsNonce($nonce, $hostname, $timestamp)) {
                        $this->error .= "The request has been used previously. ";
                    }else {
                        if(!$this->storeNonce($nonce, $hostname, $timestamp)) {
                            $this->error .="Error storing the nonces. ";
                        }else {
                            $dev = true;
                        }
                    }
                }
            }
        }
        return $dev;
    }

    private function existsNonce($nonce, $hostname, $timestamp) {
        $dev = false;
        $bd = new liteBD();
        if($bd->existsNonce($nonce, $hostname, $timestamp)) {
            $dev =true;
        }
        return $dev;
    }

    private function storeNonce($nonce, $hostname, $timestamp) {
        $dev = false;
        $bd = new liteBD();
        if($bd->storeNonce($nonce, $hostname, $timestamp)) {
            $dev =true;
        }
        return $dev;
    }


    /**
     *Function that checks if the bearer token given in the request is a valid one.
     * @param <String> $hostname IP of the client
     * @param <String> $httpsOn "on" if the connection is https
     * @return <bool>  true if the token is a valid one
     */
    private function isValidBearerToken($hostname, $httpsOn) {
        if($this->activeDebug) error_log("oauthRS.isValidBearerToken");
        return $this->isValidToken($this->bearertoken_request, $hostname, $httpsOn);
    }

    /**
     * Function that gets the Authorization header and formats the parameters on it and stores the
     *  rquests in the class parameters bearertoken_request or accesstoken_request, depending on the case
     * @param <Array> $headers Apache_authorization_headers()
     * @return <bool> True if the auth header has got a valid format.
     *
     */
    private function getAuthorizationHeader($headers) {
        if($this->activeDebug)error_log("oauthRS.getAuthorizationHeader");
        $dev = false;
        if(array_key_exists("Authorization", $headers)) {
            $value = $headers['Authorization'];
            $array = explode("Token ",$value);
            if($array == false) {
                $this->error .= "Bad Authorization Header's format - No 'Token '";
            }else {
                $array = explode(",",$array[1]);
                if($array == false) {
                    $this->error .= "Bad Authorization Header's format - No Delimiters (',')";
                }else {
                    $filtroToken='/token="(.*)"/';
                    $token = null;
                    $filtroTimestamp='/timestamp="(.*)"/';
                    $timestamp=null;
                    $filtroSignature='/signature="(.*)"/';
                    $algorithm=null;
                    $filtroAlgorithm='/algorithm="(.*)"/';
                    $signature=null;
                    $filtroNonce='/nonce="(.*)"/';
                    $nonce=null;
                    foreach ($array as $value) {
                        if(1== preg_match_all($filtroToken,$value, $out)) {
                            $token = $out[1][0];
                        }else   if(1== preg_match_all($filtroNonce,$value, $out)) {
                            $nonce = $out[1][0];
                        }else   if(1== preg_match_all($filtroAlgorithm,$value, $out)) {
                            $algorithm = $out[1][0];
                        }else  if(1== preg_match_all($filtroSignature,$value, $out)) {
                            $signature = $out[1][0];
                        }else  if(1== preg_match_all($filtroTimestamp,$value, $out)) {
                            $timestamp = $out[1][0];
                        }else {
                            $filtro='/(.*)="(.*?)"/';
                            if(1== preg_match_all($filtro,$value, $out)) {
                                array_push($this->extra, array($out[1][0] => $out[2][0]));
                            }
                        }
                    }
                    if($token==null) {
                        $this->error .=  "Bad Authorization Header's format - No token parameter";
                    }else {
                        if($signature==null || $nonce==null || $algorithm==null || $timestamp==null) {
                            $this->bearertoken_request = $token;
                            $dev=true;
                        }else {
                            $this->accesstoken_request = array("token"=>$token,"signature"=>$signature, "nonce"=>$nonce, "timestamp"=>$timestamp,"algorithm"=>$algorithm);
                            $dev=true;
                        }
                    }
                }
            }
        }
        return $dev;
    }

    /**
     * Function that gets the $_POST parameter of the request  and formats the parameters on it and stores the
     *  rquests in the class parameters bearertoken_request
     * @param <Array> $data $_POST
     * @return <bool> True if the $_POST params has got a valid format.
     *
     */
    private function getPostData($data) {
        if($this->activeDebug) error_log("oauthRS.getPostData");
        $dev = false;
        foreach ($data as $k => $val) {
            if(strcmp($k, "oauth_token")==0) {
                $this->bearertoken_request = $val;
                $dev= true;
            }else {
                $this->extra[$k] =$val;
            }
        }
        if(!$dev) {
            $this->error .= "Token format unknown in POST request. ";
        }
        return $dev;
    }

    /**
     * Function that gets the $_GET parameter of the request  and formats the parameters on it and stores the
     *  rquests in the class parameters bearertoken_request
     * @param <Array> $data $_GET
     * @return <bool> True if the $_GET params has got a valid format.
     *
     */
    private function getGetData($data) {
        if($this->activeDebug)error_log("oauthRS.getGetData");
        $dev = false;
        foreach ($data as $k => $val) {
            if(strcmp($k, "oauth_token")==0) {
                $this->bearertoken_request = $val;
                $dev= true;
            }else {
                $this->extra[$k] =$val;
            }
        }
        if(!$dev) {
            $this->error .= "Token format unknown in GET request. ";
        }
        return $dev;
    }
    /**
     * Function that returns the error to the client.
     */
    private function getMessageError() {
        if($this->activeDebug)error_log("oauthRS.getMessageError");
        $head = "HTTP/1.0 401 Unauthorized";
        if($this->isExpired) {
            $time=date("r");
            $head.=" Date: ".$time;
            $head.=" WWW-Authenticate: Token realm='service'";
            if($this->accesstoken_request!=null) {
                $head.=", algorithms='hmac-sha256', ";
            }else if($this->bearertoken_request!=null) {
                $head.=", algorithms='None', ";
            }
            $head.="error='".$this->getError()."'";
        }else {
            $head="WWW-Authenticate: OAuth";
            if($this->accesstoken_request!=null) {
                $head.=' token="'.$this->accesstoken_request['token'].'"';
            }else if($this->bearertoken_request!=null) {
                $head.=' token="'.$this->bearertoken_request.'"';
            }
        }
        header($head);
        echo "Error:  ".$this->getError();
    }

    /**
     * Function that get the keys of the clients, checking if the app client it's registered or not.
     * @param <type> $keysfile Keys archive
     */
    private function getKeys($keysfile="keys/keys.xml") {
        if($this->activeDebug) error_log("oauthRS.getKeys");
        $xml = simplexml_load_file($keysfile);
        if(strcmp($xml->getName(),"Keys")==0) {
            $arr = $xml->children();
            $aux =array();
            foreach($arr as $child) {
                $id=$child['id'];
                $val = $child['value'];
                $aux[(string)$id] = (string)$val;
            }
        }else {
            if($this->activeDebug) error_log("Bad formatting - keys.xml");
            $this->error = "Bad format of keys.xml";
        }
        $this->keys = $aux;
    }


    /**
     * Function that activates the error_log messages
     */
    private function activeDebug() {
        $this->activeDebug=true;
    }

    /**
     * Function that deactivates the error_log messages
     */
    private function deactiveDebug() {
        $this->activeDebug=false;
    }


    /* Getters*/
    public function getResource() {
        if($this->activeDebug) error_log("oauthRS.getResource");
        $res = new Resource();
        return $res->getResource();
    }

    public function getError() {
        if($this->activeDebug) error_log("oauthRS.getError");
        return $this->error;
    }

}

/**
 * Interface that must implements the class that obtains the resource.
 */
interface IResource {
    public function getResource();
}

/**
 * Resource example class
 */
class Resource implements IResource {
    function getResource() {
        $content = "<Response><ServiceProvider id='uco-consigna' protocol='papiv1' category='0'>	<Name><![CDATA[Consigna de la Universidad de Córdoba]]></Name>	<Description><![CDATA[La Consigna de la Universidad de Córdoba es un servicio para el intercambio de ficheros, donde los usuarios de SIR podrán compartir archivos incluso con usuarios que no pertenezcan a la federación.]]></Description>  <URL><![CDATA[http://consigna.uco.es/]]></URL>  <TechnicalInfo><![CDATA[http://www.rediris.es/sir/sp/consigna-uco.html]]></TechnicalInfo>	<Wayfless><![CDATA[http://sir.rediris.es/idpfirst/?idpid=aesir&spid=uco-consigna]]></Wayfless></ServiceProvider><ServiceProvider id='foodle' protocol='saml2' category='0'>	<Name><![CDATA[Foodle]]></Name>	<Description><![CDATA[Foodle es un servicio para sondeos o encuestas sencillas y para planificar encuentros.]]></Description>  <URL><![CDATA[https://foodle.feide.no/]]></URL>  <TechnicalInfo><![CDATA[http://www.rediris.es/sir/sp/foodle.html]]></TechnicalInfo>	<Wayfless><![CDATA[http://sir.rediris.es/idpfirst/?idpid=aesir&spid=foodle]]></Wayfless></ServiceProvider><ServiceProvider id='spaces-i2' protocol='saml2' category='0'>	<Name><![CDATA[Wiki federado de Internet2]]></Name>	<Description><![CDATA[El Wiki federado de Internet2 es un espacio de colaboración para actividades y grupos coordinados por la organización Internet2.]]></Description>  <URL><![CDATA[http://spaces.internet2.edu/]]></URL>  <TechnicalInfo><![CDATA[http://www.rediris.es/sir/sp/spaces-i2.html]]></TechnicalInfo>	<Wayfless><![CDATA[http://sir.rediris.es/idpfirst/?idpid=aesir&spid=spaces-i2]]></Wayfless></ServiceProvider><ServiceProvider id='msdn-aa' protocol='msdn-aa' category='1'>	<Name><![CDATA[MSDN Academic Alliance]]></Name>	<Description><![CDATA[Microsoft ofrece en su servicio MSDN Academic Alliance, un portal personalizado para cada universidad española desde el cual los estudiantes y profesores de titulaciones científico-técnicas podrán, a través de la federación SIR, descargarse todos sus productos software con licencia educativa, con la única salvedad de Office y Outlook.]]></Description>  <URL><![CDATA[http://msdn.microsoft.com/es-es/dd350178.aspx]]></URL>  <TechnicalInfo><![CDATA[http://www.rediris.es/sir/sp/msdnaa.html]]></TechnicalInfo>	<Wayfless><![CDATA[http://sir.rediris.es/idpfirst/?idpid=aesir&spid=msdn-aa]]></Wayfless></ServiceProvider><ServiceProvider id='dreamspark' protocol='shib13' category='1'>	<Name><![CDATA[Microsoft DreamSpark]]></Name>	<Description><![CDATA[DreamSpark es un programa a cargo de Microsoft para proveer a los estudiantes con la gran mayoría de sus productos con licencia gratuita educativa.]]></Description>  <URL><![CDATA[https://www.dreamspark.com]]></URL>  <TechnicalInfo><![CDATA[http://www.rediris.es/sir/sp/dreamspark.html]]></TechnicalInfo></ServiceProvider><ServiceProvider id='terena-refeds' protocol='saml2' category='2'>	<Name><![CDATA[REFEDs Wiki]]></Name>	<Description><![CDATA[]]></Description>  <URL><![CDATA[http://refeds.terena.org/]]></URL>  <TechnicalInfo><![CDATA[]]></TechnicalInfo>	<Wayfless><![CDATA[http://sir.rediris.es/idpfirst/?idpid=aesir&spid=terena-refeds]]></Wayfless></ServiceProvider><ServiceProvider id='rediris-irisrbl' protocol='papiv1' category='3'>	<Name><![CDATA[IRISRBL]]></Name>	<Description><![CDATA[Con esta herramienta, las instituciones afiliadas pueden eliminar una dirección IP de la lista que genera el servicio de reputación de RedIRIS: IRISRBL.]]></Description>  <URL><![CDATA[http://www.rediris.es/irisrbl/del/]]></URL>  <TechnicalInfo><![CDATA[http://www.rediris.es/sir/sp/irisrbl.html]]></TechnicalInfo></ServiceProvider><ServiceProvider id='atlases' protocol='shib13' category='4'>	<Name><![CDATA[Atlas of Dermatopathology]]></Name>	<Description><![CDATA[El Atlas de Dermatopatología (1997) contiene más de 3.800 imágenes, tanto clínicas como histológicas. Estas últimas están disponibles en una resolución muy alta gracias al uso de un microscopio virtual que permite acceder a ellas.]]></Description>  <URL><![CDATA[https://atlases.muni.cz/]]></URL>  <TechnicalInfo><![CDATA[http://www.rediris.es/sir/sp/atlases.html]]></TechnicalInfo>	<Wayfless><![CDATA[http://sir.rediris.es/idpfirst/?idpid=aesir&spid=atlases]]></Wayfless></ServiceProvider></Response>";
        return $content;
    }
}

?>